@@ -15,21 +15,14 @@ struct krt_proto;
 struct kif_config;
 struct kif_proto;
 
-#include "lib/krt-sys.h"
-
-/* Flags stored in net->n.flags, rest are in nest/route.h */
-
-#define KRF_VERDICT_MASK 0x0f
-#define KRF_CREATE 0			/* Not seen in kernel table */
-#define KRF_SEEN 1			/* Seen in kernel table during last scan */
-#define KRF_UPDATE 2			/* Need to update this entry */
-#define KRF_DELETE 3			/* Should be deleted */
-#define KRF_IGNORE 4			/* To be ignored */
+#include "nest/iface.h"
+#include "sysdep/config.h"
+#include CONFIG_INCLUDE_KRTSYS_H
 
 #define KRT_DEFAULT_ECMP_LIMIT	16
 
-#define EA_KRT_SOURCE	EA_CODE(EAP_KRT, 0)
-#define EA_KRT_METRIC	EA_CODE(EAP_KRT, 1)
+#define EA_KRT_SOURCE	EA_CODE(PROTOCOL_KERNEL, 0)
+#define EA_KRT_METRIC	EA_CODE(PROTOCOL_KERNEL, 1)
 
 /* Whenever we recognize our own routes, we allow learing of foreign routes */
 
@@ -44,10 +37,9 @@ extern struct protocol proto_unix_kernel;
 struct krt_config {
   struct proto_config c;
   struct krt_params sys;	/* Sysdep params */
+  btime scan_time;		/* How often we re-scan routes */
   int persist;			/* Keep routes when we exit */
-  int scan_time;		/* How often we re-scan routes */
   int learn;			/* Learn routes from other sources */
-  int devroutes;		/* Allow export of device routes */
   int graceful_restart;		/* Regard graceful restart recovery */
   int merge_paths;		/* Exported routes are merged for ECMP */
 };
@@ -57,14 +49,17 @@ struct krt_proto {
   struct krt_state sys;		/* Sysdep state */
 
 #ifdef KRT_ALLOW_LEARN
-  struct rtable krt_table;	/* Internal table of inherited routes */
+  struct rtable *krt_table;	/* Internal table of inherited routes */
 #endif
 
 #ifndef CONFIG_ALL_TABLES_AT_ONCE
   timer *scan_timer;
 #endif
 
+  struct bmap sync_map;		/* Keeps track which exported routes were successfully written to kernel */
+  struct bmap seen_map;		/* Routes seen during last periodic scan */
   node krt_node;		/* Node in krt_proto_list */
+  byte af;			/* Kernel address family (AF_*) */
   byte ready;			/* Initial feed has been finished */
   byte initialized;		/* First scan has been finished */
   byte reload;			/* Next scan is doing reload */
@@ -84,6 +79,14 @@ void kif_request_scan(void);
 void krt_got_route(struct krt_proto *p, struct rte *e);
 void krt_got_route_async(struct krt_proto *p, struct rte *e, int new);
 
+static inline int
+krt_get_sync_error(struct krt_proto *p, struct rte *e)
+{
+  return (p->p.proto_state == PS_UP) &&
+    bmap_test(&p->p.main_channel->export_map, e->id) &&
+    !bmap_test(&p->sync_map, e->id);
+}
+
 /* Values for rte->u.krt_sync.src */
 #define KRT_SRC_UNKNOWN	-1	/* Nobody knows */
 #define KRT_SRC_BIRD	 0	/* Our route (not passed in async mode) */
@@ -93,18 +96,20 @@ void krt_got_route_async(struct krt_proto *p, struct rte *e, int new);
 
 extern struct protocol proto_unix_iface;
 
-struct kif_primary_item {
-  node n;
-  byte *pattern;
-  ip_addr prefix;
-  int pxlen;
-};
-
 struct kif_config {
   struct proto_config c;
   struct kif_params sys;	/* Sysdep params */
-  int scan_time;		/* How often we re-scan interfaces */
-  list primary;			/* Preferences for primary addresses (struct kif_primary_item) */
+
+  list iface_list;		/* List of iface configs (struct kif_iface_config) */
+  btime scan_time;		/* How often we re-scan interfaces */
+};
+
+struct kif_iface_config {
+  struct iface_patt i;
+
+  ip_addr pref_v4;
+  ip_addr pref_v6;
+  ip_addr pref_ll;
 };
 
 struct kif_proto {
@@ -112,10 +117,11 @@ struct kif_proto {
   struct kif_state sys;		/* Sysdep state */
 };
 
-struct kif_proto *kif_proto;
+extern struct kif_proto *kif_proto;
 
 #define KIF_CF ((struct kif_config *)p->p.cf)
 
+struct kif_iface_config * kif_get_iface_config(struct iface *iface);
 struct proto_config * krt_init_config(int class);
 
 
@@ -134,8 +140,8 @@ void krt_sys_copy_config(struct krt_config *, struct krt_config *);
 
 int  krt_capable(rte *e);
 void krt_do_scan(struct krt_proto *);
-void krt_replace_rte(struct krt_proto *p, net *n, rte *new, rte *old, struct ea_list *eattrs);
-int krt_sys_get_attr(eattr *a, byte *buf, int buflen);
+void krt_replace_rte(struct krt_proto *p, net *n, rte *new, rte *old);
+int krt_sys_get_attr(const eattr *a, byte *buf, int buflen);
 
 
 /* kif sysdep */
@@ -150,6 +156,6 @@ void kif_sys_copy_config(struct kif_config *, struct kif_config *);
 
 void kif_do_scan(struct kif_proto *);
 
-struct ifa *kif_get_primary_ip(struct iface *i);
+int kif_update_sysdep_addr(struct iface *i);
 
 #endif
